/* SPDX-FileCopyrightText: 2026 Nikolay Govorov <me@govorov.online> */
/* SPDX-License-Identifier: AGPL-3.0-or-later */

@font-face {
  font-family: 'JetBrains Mono';
  font-style: normal;
  font-weight: 100 700;
  font-stretch: 100%;
  font-display: swap;
  src: url(./jetbrainsmono/JetBrainsMono[wght].woff2) format('woff2');
}

@font-face {
  font-family: 'JetBrains Mono';
  font-style: italic;
  font-weight: 100 700;
  font-stretch: 100%;
  font-display: swap;
  src: url(./jetbrainsmono/JetBrainsMono-Italic[wght].woff2) format('woff2');
}

*, *::before, *::after {
  box-sizing: border-box;
}

:root {
  line-height: 1.4;
  font-family: 'JetBrains Mono', monospace;
}

html {
  margin: 0;
  padding: 0;
}

body {
  margin: 0 auto;
  max-width: 960px;
  padding: 1.5em 2em;
}

code {
  margin: 0;
  padding: .2em .4em;
  white-space: break-spaces;
  background-color: #818b981f;
}

h1, h2, h3, h4, h5, h6 {
  margin: 1rem 0;
  font-weight: 700;
  line-height: 1.2;
  text-wrap: balance;
}

h1 {
  font-size: 2.75rem;
}

h1:first-child {
  margin-top: 0;
}

p, details {
  margin: 1em 0;
  text-wrap: pretty;
}

table {
  border-collapse: collapse;
}

th, td {
  padding: 7px 8px;
  border: 2px solid black;
  text-align: start;
  vertical-align: top;
}

h3 a {
  color: inherit;
  text-decoration: none;
}

h3 a:hover {
  text-decoration: underline;
}
